"""Test read only storage."""

from unittest.mock import MagicMock

import pytest
from portia import LocalDataValue, Plan, PlanRun
from portia.portia import EndUser, InMemoryStorage
from portia.storage import ToolCallRecord, ToolCallStatus

from steelthread.portia.storage import ReadOnlyStorage
from tests.unit.utils import get_test_plan_run


@pytest.fixture
def readonly_storage() -> tuple[ReadOnlyStorage, MagicMock, Plan, PlanRun]:
    """Set up storage."""
    plan, plan_run = get_test_plan_run()

    real_storage = MagicMock()
    ro = ReadOnlyStorage(storage=real_storage)
    ro.local_storage = InMemoryStorage()

    ro.local_storage.save_plan(plan)
    ro.local_storage.save_plan_run(plan_run)

    return ro, real_storage, plan, plan_run


def test_get_and_exists_fallback(
    readonly_storage: tuple[ReadOnlyStorage, MagicMock, Plan, PlanRun],
) -> None:
    """Test get and exists."""
    ro, real_storage, plan, plan_run = readonly_storage

    assert ro.get_plan(plan.id) == plan
    assert ro.get_plan_run(plan_run.id) == plan_run
    assert ro.plan_exists(plan.id) is True

    assert ro.get_plan_runs().results == [plan_run]


def test_output_and_tool_call(
    readonly_storage: tuple[ReadOnlyStorage, MagicMock, Plan, PlanRun],
) -> None:
    """Test outputs + tools."""
    ro, _, _, plan_run = readonly_storage

    output = LocalDataValue(value="foo")
    ro.save_plan_run_output("out1", output, plan_run.id)
    assert ro.get_plan_run_output("out1", plan_run.id) == LocalDataValue(value="foo")

    ro.save_tool_call(
        ToolCallRecord(
            tool_name="weather",
            plan_run_id=plan_run.id,
            step=plan_run.current_step_index,
            end_user_id="",
            status=ToolCallStatus.SUCCESS,
            input={},
            output={},
            latency_seconds=1,
        )
    )


def test_similar_plan_proxy(
    readonly_storage: tuple[ReadOnlyStorage, MagicMock, Plan, PlanRun],
) -> None:
    """Test similar plan."""
    ro, real_storage, *_ = readonly_storage
    real_storage.get_similar_plans.return_value = ["dummy"]
    assert ro.get_similar_plans("query") == ["dummy"]
    real_storage.get_similar_plans.assert_called_once()


def test_end_user_proxy(readonly_storage: tuple[ReadOnlyStorage, MagicMock, Plan, PlanRun]) -> None:
    """Test end user."""
    ro, real_storage, *_ = readonly_storage
    user = EndUser(external_id="u1", additional_data={"x": "y"})
    ro.save_end_user(user)
    assert ro.get_end_user("u1") == user


def test_fallback_on_exception(
    readonly_storage: tuple[ReadOnlyStorage, MagicMock, Plan, PlanRun],
) -> None:
    """Test error fallback."""
    ro, real_storage, _, _ = readonly_storage
    ro.local_storage.get_plan_by_query = MagicMock(side_effect=Exception("fail"))
    real_storage.get_plan_by_query.return_value = "fallback"
    assert ro.get_plan_by_query("q") == "fallback"


def test_get_plan_by_query_fallback(
    readonly_storage: tuple[ReadOnlyStorage, MagicMock, Plan, PlanRun],
) -> None:
    """Test get plan by query."""
    ro, real_storage, *_ = readonly_storage
    ro.local_storage.get_plan_by_query = MagicMock(side_effect=Exception("fail"))
    real_storage.get_plan_by_query.return_value = "fallback"
    assert ro.get_plan_by_query("q") == "fallback"


def test_plan_exists_fallback(
    readonly_storage: tuple[ReadOnlyStorage, MagicMock, Plan, PlanRun],
) -> None:
    """Test plan exists."""
    ro, real_storage, plan, *_ = readonly_storage
    ro.local_storage.plan_exists = MagicMock(side_effect=Exception("fail"))
    real_storage.plan_exists.return_value = True
    assert ro.plan_exists(plan.id) is True


def test_get_plan_runs_fallback(
    readonly_storage: tuple[ReadOnlyStorage, MagicMock, Plan, PlanRun],
) -> None:
    """Test get plan runs error."""
    ro, real_storage, *_ = readonly_storage
    ro.local_storage.get_plan_runs = MagicMock(side_effect=Exception("fail"))
    real_storage.get_plan_runs.return_value = MagicMock()
    assert ro.get_plan_runs() == real_storage.get_plan_runs.return_value


def test_get_end_user_fallback(
    readonly_storage: tuple[ReadOnlyStorage, MagicMock, Plan, PlanRun],
) -> None:
    """Test get end user."""
    ro, real_storage, *_ = readonly_storage
    ro.local_storage.get_end_user = MagicMock(side_effect=Exception("fail"))
    real_storage.get_end_user.return_value = EndUser(external_id="fallback")
    end_user = ro.get_end_user("some-id")
    assert end_user
    assert end_user.external_id == "fallback"


def test_save_plan_delegates_to_local_storage(
    readonly_storage: tuple[ReadOnlyStorage, MagicMock, Plan, PlanRun],
) -> None:
    """Test save plan."""
    ro, _, plan, _ = readonly_storage

    # Save the plan again
    ro.save_plan(plan)

    # Should now be in local_storage
    assert ro.local_storage.get_plan(plan.id) == plan


def test_get_plan_falls_back_to_storage(
    readonly_storage: tuple[ReadOnlyStorage, MagicMock, Plan, PlanRun],
) -> None:
    """Test save plan fallback."""
    ro, real_storage, plan, _ = readonly_storage

    # Force local_storage to raise
    ro.local_storage.get_plan = MagicMock(side_effect=Exception("boom"))
    # Mock fallback return
    real_storage.get_plan.return_value = plan

    result = ro.get_plan(plan.id)

    assert result == plan
    real_storage.get_plan.assert_called_once_with(plan.id)


def test_save_plan_run_delegates_to_local_storage(
    readonly_storage: tuple[ReadOnlyStorage, MagicMock, Plan, PlanRun],
) -> None:
    """Test save plan delegate."""
    ro, _, _, plan_run = readonly_storage

    # Save the plan run again
    ro.save_plan_run(plan_run)

    # Should now be in local_storage
    retrieved = ro.local_storage.get_plan_run(plan_run.id)
    assert retrieved == plan_run


def test_get_plan_run_falls_back_to_storage(
    readonly_storage: tuple[ReadOnlyStorage, MagicMock, Plan, PlanRun],
) -> None:
    """Test get plan delegate."""
    ro, real_storage, _, plan_run = readonly_storage

    # Cause local_storage to raise
    ro.local_storage.get_plan_run = MagicMock(side_effect=Exception("fail"))
    real_storage.get_plan_run.return_value = plan_run

    result = ro.get_plan_run(plan_run.id)

    assert result == plan_run
    real_storage.get_plan_run.assert_called_once_with(plan_run.id)



"""Test tool stubs."""

from typing import Never
from unittest.mock import MagicMock

import pytest
from portia import Clarification, ClarificationCategory, ToolRunContext
from portia.portia import EndUser
from portia.tool import Tool

from steelthread.portia.tools import ToolStub, ToolStubRegistry
from tests.unit.utils import get_test_config, get_test_plan_run


@pytest.fixture
def dummy_context() -> ToolRunContext:
    """Get context."""
    plan, plan_run = get_test_plan_run()
    return ToolRunContext(
        plan=plan,
        plan_run=plan_run,
        config=get_test_config(),
        clarifications=[],
        end_user=EndUser(external_id="user-123"),
    )


def test_tool_stub_with_return_callable(dummy_context: ToolRunContext) -> None:
    """Test with callable."""

    def return_callable(index, ctx, args, kwargs) -> str:  # noqa: ANN001, ARG001
        return "result"

    tool = ToolStub(
        id="test-tool",
        name="Test Tool",
        description="Test",
        output_schema=("any", "any"),
        return_callable=return_callable,
        tool_calls=[],
    )

    assert tool.ready(dummy_context).ready
    result = tool.run(dummy_context, 1, x=2)
    assert result == "result"
    assert len(tool.tool_calls) == 1
    assert tool.tool_calls[0].status.name == "SUCCESS"


def test_tool_stub_with_child_tool(dummy_context: ToolRunContext) -> None:
    """Test with child tool."""

    class DummyChildTool(Tool):
        def run(self, ctx, *args, **kwargs) -> str:  # noqa: ANN001, ANN002, ANN003, ARG002
            return "child-result"

    child = DummyChildTool(
        id="child",
        name="Child Tool",
        description="",
        output_schema=("any", "any"),
    )
    tool = ToolStub(
        id="stub",
        name="Stub Tool",
        description="",
        output_schema=("any", "any"),
        child_tool=child,
        tool_calls=[],
    )

    result = tool.run(dummy_context)
    assert result == "child-result"
    assert tool.tool_calls[0].status.name == "SUCCESS"


def test_tool_stub_with_child_tool_error(dummy_context: ToolRunContext) -> None:
    """Test tool child fail."""

    class DummyChildTool(Tool):
        def run(self, ctx, *args, **kwargs) -> Never:  # noqa: ANN001, ANN002, ANN003, ARG002
            raise ValueError("err from child tool")

    child = DummyChildTool(
        id="child",
        name="Child Tool",
        description="",
        output_schema=("any", "any"),
    )
    tool = ToolStub(
        id="stub",
        name="Stub Tool",
        description="",
        output_schema=("any", "any"),
        child_tool=child,
        tool_calls=[],
    )

    result = tool.run(dummy_context)
    assert result == "err from child tool"
    assert tool.tool_calls[0].status.name == "FAILED"


def test_tool_stub_failure_from_callable(dummy_context: ToolRunContext) -> None:
    """Test fail in callable."""

    def failing_callable(index, ctx, args, kwargs) -> Never:  # noqa: ANN001, ARG001
        raise RuntimeError("fail")

    tool = ToolStub(
        id="fail-tool",
        name="Fail Tool",
        description="",
        output_schema=("any", "any"),
        return_callable=failing_callable,
        tool_calls=[],
    )

    result = tool.run(dummy_context)
    assert "fail" in result
    assert tool.tool_calls[0].status.name == "FAILED"


def test_tool_stub_fails_without_child_or_callable(dummy_context: MagicMock) -> None:
    """Test fails."""
    tool = ToolStub(
        id="bad-tool",
        name="Bad Tool",
        description="",
        output_schema=("any", "any"),
        tool_calls=[],
    )

    with pytest.raises(RuntimeError, match="must have either child_tool or return_callable"):
        tool.run(dummy_context)


def test_tool_stub_sets_plan_run_id_on_clarification(dummy_context: MagicMock) -> None:
    """Test stub w clarification."""

    def returns_clarification(index, ctx, args, kwargs) -> Clarification:  # noqa: ANN001, ARG001
        return Clarification(
            category=ClarificationCategory.INPUT,
            user_guidance="help",
            plan_run_id=ctx.plan_run.id,
        )

    tool = ToolStub(
        id="clarify-tool",
        name="Clarify Tool",
        description="",
        output_schema=("any", "any"),
        return_callable=returns_clarification,
        tool_calls=[],
    )

    result = tool.run(dummy_context)
    assert isinstance(result, Clarification)
    assert result.plan_run_id == dummy_context.plan_run.id


def test_tool_stub_registry_resolves_stubs() -> None:
    """Test resolving stubs."""

    class DummyChildTool(Tool):
        def run(self, ctx, *args, **kwargs) -> str:  # noqa: ANN001, ANN002, ANN003, ARG002
            return "child-result"

    tool = DummyChildTool(
        id="my-tool",
        name="Tool",
        description="desc",
        output_schema=("any", "any"),
    )

    registry = MagicMock()
    registry.get_tools.return_value = [tool]

    def stub_fn(index, ctx, args, kwargs) -> str:  # noqa: ANN001, ARG001
        return "ok"

    stub_registry = ToolStubRegistry(registry=registry, stubs={"my-tool": stub_fn})
    resolved = stub_registry.get_tool("my-tool")
    assert isinstance(resolved, ToolStub)
    assert resolved.return_callable is stub_fn


def test_tool_stub_registry_fallbacks_and_calls_tracking() -> None:
    """Tool that is not explicitly stubbed."""

    class DummyChildTool(Tool):
        def run(self, ctx, *args, **kwargs) -> str:  # noqa: ANN001, ANN002, ANN003, ARG002
            return "child-result"

    base_tool = DummyChildTool(
        id="base-tool",
        name="Tool",
        description="desc",
        output_schema=("any", "any"),
    )
    registry = MagicMock()
    registry.get_tools.return_value = [base_tool]
    registry.get_tool.return_value = base_tool

    stub_registry = ToolStubRegistry(registry=registry, stubs={})

    resolved = stub_registry.get_tool("base-tool")
    assert isinstance(resolved, ToolStub)
    assert resolved.child_tool == base_tool

    calls = stub_registry.get_tool_calls("base-tool")
    assert calls == []

    calls = stub_registry.get_tool_calls("other-tool")
    assert calls == []

    all_calls = stub_registry.get_tool_calls()
    assert isinstance(all_calls, list)

    def stub_response(index: int, ctx: ToolRunContext, args: tuple, kwargs: dict) -> str:  # noqa: ARG001
        return "stub"

    stub_registry = ToolStubRegistry(registry=registry, stubs={"base-tool": stub_response})

    # should cache tool stubs
    resolved1 = stub_registry.get_tool("base-tool")
    assert isinstance(resolved, ToolStub)
    resolved2 = stub_registry.get_tool("base-tool")
    assert isinstance(resolved, ToolStub)
    assert resolved1 == resolved2

    tools = stub_registry.get_tools()
    assert isinstance(tools, list)



"""Test backend."""

from unittest.mock import MagicMock, patch

import pytest
from httpx import Response

from steelthread.offline_evaluators.test_case import OfflineTestCase
from steelthread.online_evaluators.test_case import OnlineTestCase
from steelthread.portia.backend import PortiaBackend
from tests.unit.utils import get_test_config


@pytest.fixture
def backend() -> PortiaBackend:
    """Get backend."""
    return PortiaBackend(config=get_test_config())


def make_response(status_code: int = 200, json_data: object | None = None) -> MagicMock:
    """Make a fake response."""
    mock_response = MagicMock(spec=Response)
    mock_response.is_success = status_code < 400
    mock_response.status_code = status_code
    mock_response.content = b'{"error": "bad"}'
    mock_response.json.return_value = json_data or []
    return mock_response


@patch("steelthread.portia.backend.PortiaCloudClient")
def test_client_returns_httpx_client(mock_portia_client: MagicMock, backend: PortiaBackend) -> None:
    """Check client returns client."""
    mock_client = MagicMock()
    mock_portia_client.return_value.get_client.return_value = mock_client

    client = backend.client()

    mock_portia_client.return_value.get_client.assert_called_once_with(backend.config)
    assert client == mock_client


def test_check_response_success(backend: PortiaBackend) -> None:
    """Check response 200."""
    response = make_response(status_code=200)
    backend.check_response(response)  # should not raise


def test_check_response_failure_raises(backend: PortiaBackend) -> None:
    """Check response 500."""
    response = make_response(status_code=500)
    with pytest.raises(ValueError, match="bad"):
        backend.check_response(response)


@patch.object(PortiaBackend, "client")
def test_load_online_evals(client_mock: MagicMock, backend: PortiaBackend) -> None:
    """Check load evals."""
    client = MagicMock()
    response = make_response(
        json_data=[{"id": "abc", "related_item_id": "123", "related_item_type": "plan"}]
    )
    client.get.return_value = response
    client_mock.return_value = client

    cases = backend.load_online_evals("test-dataset")

    client.get.assert_called_once_with(url="/api/v0/evals/test-cases/?dataset_name=test-dataset")
    assert isinstance(cases[0], OnlineTestCase)


@patch.object(PortiaBackend, "client")
def test_load_offline_evals(client_mock: MagicMock, backend: PortiaBackend) -> None:
    """Check load offline."""
    client = MagicMock()
    response = make_response(
        json_data=[
            {
                "id": "abc",
                "input_config": {"type": "query", "value": "do a backflip."},
                "assertions": [],
            }
        ]
    )
    client.get.return_value = response
    client_mock.return_value = client

    cases = backend.load_offline_evals("test-dataset")

    client.get.assert_called_once_with(url="/api/v0/evals/test-cases/?dataset_name=test-dataset")
    assert isinstance(cases[0], OfflineTestCase)


@patch.object(PortiaBackend, "client")
def test_mark_processed_calls_patch(client_mock: MagicMock, backend: PortiaBackend) -> None:
    """Check processed."""
    client = MagicMock()
    response = make_response()
    client.patch.return_value = response
    client_mock.return_value = client

    tc = OnlineTestCase(id="123", related_item_id="", related_item_type="123")

    backend.mark_processed(tc)

    client.patch.assert_called_once_with(
        url="/api/v0/evals/test-cases/",
        json={"processed": True, "id": "123"},
    )



"""Backend for Portia evals."""

import httpx
from portia import Config
from portia.storage import PortiaCloudClient
from pydantic import BaseModel

from steelthread.offline_evaluators.test_case import OfflineTestCase
from steelthread.online_evaluators.test_case import OnlineTestCase


class PortiaBackend(BaseModel):
    """Client interface for interacting with the Portia API for evaluations.

    Provides methods to load test cases and mark them as processed.

    Attributes:
        config (Config): The Portia configuration containing API credentials and context.

    """

    config: Config

    def client(self) -> httpx.Client:
        """Create an HTTP client for interacting with the Portia API.

        Returns:
            httpx.Client: A configured HTTP client.

        """
        return PortiaCloudClient().get_client(self.config)

    def check_response(self, response: httpx.Response) -> None:
        """Validate the response from Portia API.

        Args:
            response (httpx.Response): The response from the Portia API to check.

        Raises:
            ValueError: If the response status code indicates an error.

        """
        if not response.is_success:
            error_str = str(response.content)
            raise ValueError(error_str)

    def load_online_evals(self, data_set: str) -> list[OnlineTestCase]:
        """Load online test cases from the Portia API.

        Args:
            data_set (str): The dataset name to fetch test cases for.

        Returns:
            list[OnlineTestCase]: A list of parsed online test cases.

        """
        client = self.client()
        response = client.get(
            url=f"/api/v0/evals/test-cases/?dataset_name={data_set}",
        )
        self.check_response(response)
        response_json = response.json()
        test_cases = []
        for tc in response_json:
            ds = OnlineTestCase(**tc)
            test_cases.append(ds)
        return test_cases

    def load_offline_evals(self, data_set: str) -> list[OfflineTestCase]:
        """Load offline test cases from the Portia API.

        Args:
            data_set (str): The dataset name to fetch test cases for.

        Returns:
            list[OfflineTestCase]: A list of parsed offline test cases.

        """
        client = self.client()
        response = client.get(
            url=f"/api/v0/evals/test-cases/?dataset_name={data_set}",
        )
        self.check_response(response)
        response_json = response.json()
        test_cases = []
        for tc in response_json:
            ds = OfflineTestCase(**tc)
            test_cases.append(ds)
        return test_cases

    def mark_processed(self, tc: OnlineTestCase) -> None:
        """Mark an online test case as processed in the Portia API.

        Args:
            tc (OnlineTestCase): The test case to mark as processed.

        """
        client = self.client()
        response = client.patch(
            url="/api/v0/evals/test-cases/",
            json={"processed": True, "id": tc.id},
        )
        self.check_response(response)
